#!/usr/bin/perl -w

# Perl script for performing various rendering-related tasks
# (c) 2000 by Peter Popov

use File::Find;
use File::Basename;

# Default values of variables used throughout this script
# All of them can be changed through the command line

$Output_Directory  = "./output/";
$Thumbs_Directory  = "./thumbnails/";
$Images_Directory  = "./images/";
$Stream_Directory  = "./streams/";
$Input_Directory   = "./";
@Files_To_Render   = ();
$Input_File        = "";

$Render_Engine     = "megapov";

@Boolean = ("No", "Yes");

$JPEG = 0;
$GIF  = 1;

@Supported_Formats = ();
$Supported_Formats[$JPEG] = "JPEG";
$Supported_Formats[$GIF] = "GIF";

@Filename_Extensions = ();
$Filename_Extensions[$JPEG] = "jpg";
$Filename_Extensions[$GIF] = "gif";


$UNIX    = 0;
$MSWIN32 = 1;
$MSDOS   = 2;
$AMIGAOS = 3;
$VMS     = 4;

@Supported_OSs = ();
$Supported_OSs[$UNIX]    = "UNIX";
$Supported_OSs[$MSWIN32] = "MSWin32";
$Supported_OSs[$MSDOS]   = "MSDOS";
$Supported_OSs[$AMIGAOS] = "AmigaOS";
$Supported_OSs[$VMS]     = "VMS";

$Operating_System        = $UNIX;

$Convert_Image           = 1;
$Make_Thumbnail          = 1;
$Thumbnail_Size          = "160x120";
$Image_Format            = $JPEG;
$Jpeg_Quality            = 95;
$Gif_Colors              = 256;
$Jpeg_Thumbnail_Quality  = 50;

$Ini_File                = "./render.ini";
$Image_Magick            = "convert";

# Various options

$Verbose_Level           = 1;    #0 fo silent, 1 for normal, 2 for verbose
$Auto_Continue           = 0;
$Clean                   = 0;
$Newer_Only              = 1;
$Replicate_Tree          = 0;
$Test_Only               = 0;

# ------------------------------------------------------------------------

# List of command-line options and their indices. KEEP THEM IN SYNC!!!

@Short_Arguments = (
  "-c",             # "--convert"
  "-I",             # "--input-directory"
  "-f",             # "--ini-file"
  "-i",             # "--input-file"
  "-m",             # "--make-thumbnail"
  "-O",             # "--output-directory"
  "-t",             # "--image-format"
  "-g",             # "--gif-colors"
  "-e",             # "--render-engine"
  "-q",             # "--quiet"
  "-v",             # "--verbose"
  "-nc",            # "--no-conversion"
  "-j",             # "--jpeg-quality"
  "-T",             # "--no-thumbnail"
  "-S",             # "--thumbnail-size"
  "-td",            # "--thumbs-directory"
  "-id",            # "--images-directory"
  "-sd",            # "--stream-directory"
  "-J",             # "--jpeg-thumbnail-quality"
  "-OS",            # "--operating-system"
  "-ac",            # "--auto-continue"
  "-n",             # "--newer-only"
  "-C",             # "--clean"
  "-rt",            # "--replicate-tree"
  "-o",             # "--overwrite"
  "-test"           # "--test-only"
);

@Arguments = (
  "--convert",
  "--input-directory",
  "--ini-file",
  "--input-file",
  "--make-thumbnail",
  "--output-directory",
  "--image-format",
  "--gif-colors",
  "--render-engine",
  "--quiet",
  "--verbose",
  "--no-conversion",
  "--jpeg-quality",
  "--no-thumbnail",
  "--thumbnail-size",
  "--thumbs-directory",
  "--images-directory",
  "--stream-directory",
  "--jpeg-thumbnail-quality",
  "--operating-system",
  "--auto-continue",
  "--newer-only",
  "--clean",
  "--replicate-tree",
  "--overwrite",
  "--test-only"             
);

$Enum = -1;

$CONVERT                = ++$Enum;    
$INPUT_DIRECTORY        = ++$Enum;    
$INI_FILE               = ++$Enum;    
$INPUT_FILE             = ++$Enum;    
$MAKE_THUMBNAIL         = ++$Enum;    
$OUTPUT_DIRECTORY       = ++$Enum;    
$IMAGE_FORMAT           = ++$Enum;    
$GIF_COLORS             = ++$Enum;    
$RENDER_ENGINE          = ++$Enum;    
$QUIET                  = ++$Enum;    
$VERBOSE                = ++$Enum;     
$NO_CONVERT             = ++$Enum;
$JPEG_QUALITY           = ++$Enum;
$NO_THUMBNAIL           = ++$Enum;
$THUMBNAIL_SIZE         = ++$Enum;
$THUMBS_DIRECTORY       = ++$Enum;
$IMAGES_DIRECTORY       = ++$Enum;
$STREAM_DIRECTORY       = ++$Enum;
$JPEG_THUMBNAIL_QUALITY = ++$Enum;
$OPERATING_SYSTEM       = ++$Enum;
$AUTO_CONTINUE          = ++$Enum;
$NEWER_ONLY             = ++$Enum;
$CLEAN                  = ++$Enum;
$REPLICATE_TREE         = ++$Enum;
$OVERWRITE              = ++$Enum;
$TEST_ONLY              = ++$Enum;

#----------------------------- Subroutines -------------------------------

# ---------
#
# Matches a string to the set of predefined command-line options. Returns
# the token ID of the option that matched. Handles ambiguous or invalid
# otions
#
# Works :)
#
# ---------

sub Parse_Argument
{
  my $Matches = 0;
  
  for (my $j=0; $j < @Arguments; $j++) {
    if ( substr ( $Arguments[$j], 0, length ($_[0]) ) eq $_[0] ) { $Matches++; $Match = $j; }
  }
  
  if ($Matches == 0) {
    for ($j=0; $j < @Short_Arguments; $j++) {
      if ( $Short_Arguments[$j] eq $_[0] ) { $Matches++; $Match = $j; }
    }
  }
  
  if ($Matches == 0) { die ("\nInvalid option ", $_[0], "\n"); }
  if ($Matches > 1) { die ("\nAmbiguous option. '", $_[0], "' is not an unique prefix.\n\n"); }
  return $Match;
}

# ---------
#
# Recurses a directory and returns a list of all .pov files with pathnames.
#
# This subroutine is based on code shamelessly ripped off the manpage :)
#
# Does not explore symbolic links (yet).
#
# ---------

sub Find_All_POV_Files
{
  my $Z=0;
  find sub { $Files_To_Render[$Z++] = $File::Find::name, -d && '/' . "\n" if /\.pov$/i}, @_;
  split (@Files_To_Render, "\n");
}

sub To_Upcase
{
  my $Foo = $_[0];
  $Foo =~ tr/a-z/A-Z/;
  return $Foo;
}

sub To_Lowcase
{
  my $Foo = $_[0];
  $Foo =~ tr/a-z/A-Z/;
  return $Foo;
}

sub Err_No_Value
{
  die "Value expected after ``$Arguments[$_[0]]'' (or ``$Short_Arguments[$_[0]]'')!\n"
}


sub Err_No_Number
{
  die "The ``$Arguments[$_[0]]'' (or ``$Short_Arguments[$_[0]]'') option requires an integer value but ``$_[1]'' doesn't look like one.\n\n";
}
  
sub Err_Out_Of_Range
{
  die "The ``$Arguments[$_[0]]'' (or ``$Short_Arguments[$_[0]]'') option requires a value in the range [$_[2]..$_[3]] but $_[1] is not in it.\n\n";
}

sub Check_File  # (Filename, r, w, x, d)
{
  if (!-e $_[0]) { die "\n``$_[0]'' does not exist.\n";
    if (!($_[1] && -r $_[0])) { die "\nRead permission denied on file ``$_[0]''.\n"; }
    if (!($_[2] && -w $_[0])) { die "\nWrite permission denied on file ``$_[0]''.\n"; }
    if (!($_[3] && -x $_[0])) { die "\nExecute permission denied on file ``$_[0]''.\n"; }
    if (!($_[4] && -d $_[0])) { die "\n``$_[0]'' is not a valid directory.\n"; }
  }
}

sub Check_Directory
{ 
  @Levels = split ("/", $_[1]);
  if ( !@Levels) { splice (@Levels, 0, 0, ""); }
  for ($k = 0; $k < @Levels; $k++) {
    $Test_Dir = "";
    for ($l = 0; $l <= $k; $l++) { $Test_Dir .= $Levels[$l] . "/" if length ($Levels[$l]); }
    if ( !-e ($_[0] . $Test_Dir)) {
      if ( !$Replicate_Tree ) {
        die "Directory ``$_[0]$Test_Dir'' does not exist. Exitting.\n";
      }
      else
      {
        if ( $Verbose_Level) { print "$_[0]$Test_Dir does not exist, creating it.\n"; }
        $Shellout = "mkdir $_[0]$Test_Dir";
        if ( $Test_Only ) {
          if ($Verbose_Level == 2) {
            print $Shellout, "\n";
          }
        }
        else { system $Shellout; }
      }
    }
  }
}
    

sub Display_Extended_Info
{ 
  print "\n", "Current settings:\n";
  print "\n", "Output directory       = $Output_Directory\t For raw ppm output";
  print "\n", "Thumbnails directory   = $Thumbs_Directory\t For thumbnails";
  print "\n", "Images directory       = $Images_Directory\t For full-size compressed images";
  print "\n", "Streams directory      = $Stream_Directory\t For the various streams (debug, render...)";
  print "\n";
  print "\n", "Replicate dir tree     = $Boolean[$Replicate_Tree]\t Replicate directory structure";
  print "\n", "Input file             = $Input_File\t The file to render";
  print "\n", "Input directory        = $Input_Directory\t Directory with files for batch render";
  print "\n", "INI file               = $Ini_File\t This file will be used for all files in the batch";
  print "\n", "Auto continue render   = $Boolean[$Auto_Continue]\t Continue render if raw output file found";
  print "\n";
  print "\n", "Convert image          = $Boolean[$Convert_Image]";
  print "\n", "Image format           = $Supported_Formats[$Image_Format]\t GIF and JPEG are currently supported";
  print "\n", "JPEG quality           = $Jpeg_Quality\t [1..100] , only used if JPEG output selected";
  print "\n", "GIF colors             = $Gif_Colors\t [2..256] , only used if GIF output selected)";
  print "\n", "Make thumbnail         = $Boolean[$Make_Thumbnail]";
  print "\n", "Thumbnail size         = $Thumbnail_Size\t Use FOOxBAR format";
  print "\n", "JPEG thumbnail quality = $Jpeg_Thumbnail_Quality\t Only used if JPEG output selected";
  print "\n", "Clean after converting = $Boolean[$Clean]\t Delete raw output file after convertion";
  print "\n", "Newer files only       = $Boolean[$Newer_Only]\t Do not render file if converted image found";
  print "\n";
  print "\n", "Operating system       = $Supported_OSs[$Operating_System]\t UNIX, MSDOS, AMIGAOS, MSWIN32 or VMS";
  print "\n", "Verbose level          = $Verbose_Level\t It's 2 or higher if you're reading this :)";
  print "\n", "Test mode enabled      = $Boolean[$Test_Only]\t Debug mode. Do not use with --quiet.";
  print "\n\n";
}

sub Help
{ 
  print "Usage: render [option [<value>]]...\n\n";
  print "Use --verbose for some info on the commands and default values.\n";
  for ($Z=0; $Z<@Arguments; $Z++) {
    print $Short_Arguments[$Z], "\t", $Arguments[$Z], "\n";
  }
  print "\n";
}

#--------------------------------------------------------------------------

# Parse command line

print "A simple script for making batch renders easy.\n";
print "(C) 2000 by Peter Popov\n\n";

if (@ARGV) {
MAINLOOP:
  for (my $i=0; $i < @ARGV; $i++) {
    $Current = Parse_Argument ($ARGV[$i]);
    $Another = ($i < @ARGV-1);
    if ($Another) { $Next = $ARGV[$i+1]; }
    
    if ( $Current == $VERBOSE )          { $Verbose_Level = 2;                    next; }
    if ( $Current == $QUIET )            { $Verbose_Level = 0;                    next; }
    if ( $Current == $CONVERT )          { $Convert_Image = 1;                    next; }
    if ( $Current == $NO_CONVERT )       { $Convert_Image = 0;                    next; }
    if ( $Current == $MAKE_THUMBNAIL  )  { $Make_Thumbnail = 1;                   next; }
    if ( $Current == $NO_THUMBNAIL  )    { $Make_Thumbnail = 0;                   next; }
    if ( $Current == $AUTO_CONTINUE )    { $Auto_Continue = 1;                    next; }
    if ( $Current == $REPLICATE_TREE )   { $Replicate_Tree = 1;                   next; }
    if ( $Current == $CLEAN  )           { $Clean = 1;                            next; }
    if ( $Current == $NEWER_ONLY )       { $Newer_Only = 1;                       next; }
    if ( $Current == $OVERWRITE )        { $Newer_Only = 0;                       next; }
    if ( $Current == $TEST_ONLY )        { $Test_Only = 1;                        next; }
                                         
    if ( $Current == $RENDER_ENGINE   )  { if ($Another) {
                                             $Render_Engine = $Next;
                                             Check_File ($Render_Engine,1,0,1,0);
                                           }
                                         }
                                         
    if ( $Current == $INPUT_FILE )       { if ($Another) {
                                             $Input_File = $Next;
                                             Check_File ($Input_File,1,0,0,0);
                                             ($Foo,$Input_Directory,$Bar) = fileparse ($Input_File,'\.[pP][oO][vV]');
                                             $Foo=$Bar;  # make -w happy
                                             @Files_To_Render=($Input_File);
                                             $i++; next;
                                           }
                                           else { Err_No_Value ($Current); }
                                         }
                                         
    if ( $Current == $GIF_COLORS      )  { if ($Another) {
                                             $Gif_Colors = $Next;
                                             if ($Gif_Colors =~ /^\d+$/) {
                                               if ($Gif_Colors >= 2 && $Gif_Colors <= 256) {
                                                 $i++;  next;
                                               }
                                               else { Err_Out_Of_Range ($Current, $Gif_Colors, 2, 256); }
                                             }
                                             else { Err_No_Number ($Current, $Gif_Colors); }
                                           }
                                           else { Err_No_Value ($Current); }
                                         }
                                         
    if ( $Current == $INPUT_DIRECTORY )  { if ($Another) {
                                             $Input_Directory = $Next;
                                             Check_File ($Input_Directory, 1, 1, 0, 1);
                                             Find_All_POV_Files ($Input_Directory);
#                                             if ( substr ($Input_Directory, -1) ne "/" ) {
#                                               $Input_Directory .= "/";
#                                             }
                                             $Input_File = "";
                                             $i++;  next;
                                           }
                                           else { Err_No_Value ($Current); }
                                         }
                                          
    if ( $Current == $OUTPUT_DIRECTORY ) { if ($Another) {
                                             $Output_Directory = $Next;
                                             Check_File ($Output_Directory, 1, 1, 0, 1);
                                             $i++;  next;
                                           }
                                           else { Err_No_Value ($Current); }
                                         }
                                          
    if ( $Current == $JPEG_QUALITY )     { if ($Another) {
                                             $Jpeg_Quality = $Next; 
                                             if ($Jpeg_Quality =~ /^\d+$/) {
                                               if ($Jpeg_Quality >= 0 && $Jpeg_Quality <= 100) {
                                                 $i++; next;
                                               }
                                               else { Err_Out_Of_Range ($Current, $Jpeg_Quality, 0, 100) }
                                             }
                                             else { Err_No_Number ($Current, $Jpeg_Quality) }
                                           }
                                           else { Err_No_Value ($Current); }
                                         }
                                         
    if ( $Current == $JPEG_THUMBNAIL_QUALITY )
                                         { if ($Another) {
                                             $Jpeg_Thumbnail_Quality = $Next; 
                                             if ($Jpeg_Thumbnail_Quality =~ /^\d+$/) {
                                               if ($Jpeg_Thumbnail_Quality >= 0 && $Jpeg_Thumbnail_Quality <= 100) {
                                                 $i++; next;
                                               }
                                               else { Err_Out_Of_Range ($Current, $Jpeg_Thumbnail_Quality, 0, 100) }
                                             }
                                             else { Err_No_Number ($Current, $Jpeg_Thumbnail_Quality) }
                                           }
                                           else { Err_No_Value ($Current); }
                                         }
                                         
    if ( $Current == $IMAGE_FORMAT )     { if ($Another) {
                                             if ($Next eq $JPEG || To_Upcase ($Next) eq "JPEG") {
                                               $Image_Format=$JPEG; $i++; next; }
                                               elsif ($Next eq $GIF || To_Upcase ($Next) eq "GIF") {
                                                 $Image_Format=$GIF; $i++; next; }
                                                 else { die "``$Next'' is not a valid value for the ``$Arguments[$Current]'' option.\n\n"; }
                                           }
                                           else { Err_No_Value ($Current); }
                                         }
                                         
    if ( $Current == $INI_FILE )         { if ($Another) {
                                             $Ini_File = $Next;
                                             Check_File ($Ini_File, 1, 0, 0, 0);
                                             $i++; next;
                                           }
                                           else { Err_No_Value ($Current); }
                                         }
                                         
    if ( $Current == $THUMBNAIL_SIZE )   { if ($Another) {
                                             $Thumbnail_Size = $Next;
                                             if ($Thumbnail_Size =~ /^\d+x\d+$/) {
                                               $i++; next;
                                             }
                                             else {
                                               die "Invalid value for ``$Arguments[$Current]'' .\n"
                                             }
                                           }
                                         }
                                         
    if ( $Current == $THUMBS_DIRECTORY)  { if ($Another) {
                                             $Thumbs_Directory = $Next;
                                             Check_File($Thumbs_Directory,1,1,0,1);
                                             $i++; next;
                                           }
                                           else { Err_No_Value ($Current) }
                                         }
                                         
    if ( $Current == $IMAGES_DIRECTORY)  { if ($Another) {
                                             $Images_Directory = $Next;
                                             Check_File ($Images_Directory,1,1,0,1);
                                             $i++; next;
                                           }
                                           else { Err_No_Value ($Current) }
                                         }
                                         
    if ( $Current == $STREAM_DIRECTORY)  { if ($Another) {
                                             $Stream_Directory = $Next;
                                             Check_File($Stream_Directory,1,1,0,1);
                                             $i++; next;
                                           }
                                           else { Err_No_Value ($Current) }
                                         }
                                         
    if ( $Current == $OPERATING_SYSTEM ) { if ($Another) {
                                             if (To_Upcase ($Next) eq "VMS") { $Operating_System=$VMS; $i++; next MAINLOOP; }
                                               elsif (To_Upcase ($Next) eq "MSDOS") { $Operating_System=$MSDOS; $i++; next MAINLOOP; }
                                               elsif (To_Upcase ($Next) eq "MSWIN32") { $Operating_System=$MSWIN32; $i++; next MAINLOOP; }
                                               elsif (To_Upcase ($Next) eq "AMIGAOS") { $Operating_System=$AMIGAOS; $i++; next MAINLOOP; }
                                               elsif ((To_Upcase ($Next) eq "UNIX") || (To_Upcase ($Next) eq "LINUX")) { $Operating_System = $UNIX; $i++; next MAINLOOP; }
                                               else { print "``$ARGV[$i+1]'' is not a supported OS, UNIX assumed.\n";
                                               $Operating_System=$UNIX; $i++; next MAINLOOP;
                                             }
                                           }
                                           else { Err_No_Value ($Current); }
                                         }

  }
}
else { Help }

if ($Verbose_Level == 2) { Display_Extended_Info }

Check_Directory ($Output_Directory, "");
Check_Directory ($Stream_Directory, "");
    
if ( $Convert_Image ) { Check_Directory ( $Images_Directory, ""); }
if ( $Make_Thumbnail ) { Check_Directory ( $Thumbs_Directory, ""); }

    
foreach (@Files_To_Render) {
  ($Stripped_File_Name, $Current_File_Directory, $Suffix) = fileparse($_, '\.[pP][oO][vV]');
  
  $Render_This_File = 1;
  $Subdirectory = "";
  
  if ( $Replicate_Tree ) {
    $Current_File_Directory =~ /$Input_Directory(\/)?/;
    $Subdirectory = $';
    
    Check_Directory ($Output_Directory, $Subdirectory);
    Check_Directory ($Stream_Directory, $Subdirectory);
    
    if ( $Convert_Image ) { Check_Directory ( $Images_Directory, $Subdirectory); }
    if ( $Make_Thumbnail ) { Check_Directory ( $Thumbs_Directory, $Subdirectory); }
  }

  if ( $Newer_Only ) {
    if (-e ($Images_Directory . $Subdirectory . $Stripped_File_Name . "." . $Filename_Extensions[$Image_Format])) {
      print "File ", $Input_Directory, $Subdirectory, $Stripped_File_Name, $Suffix, " has already been rendered. Skipped.\n";
      $Render_This_File = 0;
    }
  }
  
  if ( $Render_This_File ) {
    $Shellout = $Render_Engine . " " . $Ini_File;
    if ( -e "$Input_Directory$Subdirectory$Stripped_File_Name.ini" ) {
      $Shellout .= " $Input_Directory$Subdirectory$Stripped_File_Name.ini";
      if ($Verbose_Level) {
        print "Found $Input_Directory$Subdirectory$Stripped_File_Name.ini, using it.\n";
      }
    }
    
    $Shellout .=
      " -i"  . $_ .
      " -o"  . $Output_Directory . $Subdirectory .
      " -l"  . $Current_File_Directory .
      " -GR" . $Stream_Directory . $Subdirectory . $Stripped_File_Name . ".render" .
      " -GF" . $Stream_Directory . $Subdirectory . $Stripped_File_Name . ".fatal" .
      " -GD" . $Stream_Directory . $Subdirectory . $Stripped_File_Name . ".debug" .
      " -GW" . $Stream_Directory . $Subdirectory . $Stripped_File_Name . ".warning" .
      " -GS" . $Stream_Directory . $Subdirectory . $Stripped_File_Name . ".statistics" .
      " 2> " . $Stream_Directory . $Subdirectory . $Stripped_File_Name . ".status"
      ;
    
    if ($Auto_Continue) {
      if (-e ($Output_Directory . $Stripped_File_Name . ".ppm")) {
        if ($Verbose_Level) {
          print "File ", $Output_Directory . $Subdirectory . $Stripped_File_Name . ".ppm" .
                " exists, forcing continued render.\n";
        }
      }
    }
    
    if ($Verbose_Level) { print "Now rendering $_... \n"; }
    
     if ($Test_Only) { if ($Verbose_Level == 2) { print $Shellout, "\n"; } }
     else {
       $POV_Exit_Status = system ($Shellout) / 256;
     }
     
     if ( $POV_Exit_Status ) {
       if ( $POV_Exit_Status == 1) {
         print "Rendered exitted with an error.\n";
         print "Please review $Stream_Directory$Subdirectory$Stripped_File_Name.fatal for details.\n";
       }
       elsif ( $POV_Exit_Status == 2 ) {
         print "User abort. ";
         if ( $Clean ) { print "Output file will NOT be deleted. "; }
         if ( $Convert_Image || $Make_Thumbnail ) { print "Image conversion process skipped.\n"; }
       }
       print "\n";
     }
     
     else {
       if ( $Convert_Image) {
         $Shellout =
           $Image_Magick .
           " -compress " .
           $Supported_Formats[$Image_Format]
         ;
         if ($Image_Format == $JPEG) {
           $Shellout .= " -quality $Jpeg_Quality ";
         }  
         if ($Image_Format == $GIF) {
           $Shellout .= " -dither " . "-colors $Gif_Colors ";
         }
         $Shellout .= " $Output_Directory" . $Subdirectory . $Stripped_File_Name . ".ppm ";
         if ($Image_Format == $JPEG) {
           $Shellout .= " $Images_Directory" . $Subdirectory . $Stripped_File_Name . ".jpg ";
         }
         else {
           $Shellout .= " $Images_Directory" . $Subdirectory . $Stripped_File_Name . ".gif ";
         }
        
        if ($Verbose_Level) { print "Converting to $Supported_Formats[$Image_Format] format... \n"; }
        
        if ($Test_Only) { if ($Verbose_Level == 2) { print $Shellout, "\n"; } }
        else { system ($Shellout); }
     
        if ($Verbose_Level) { print "\n"; }
     
      }
    
      if ($Make_Thumbnail) {
        $Shellout =
          $Image_Magick .
          " -antialias" .
          " -compress " .
          $Supported_Formats[$Image_Format] .
          " -geometry " . $Thumbnail_Size
        ;
        if ($Image_Format == $JPEG) {
          $Shellout .= " -quality $Jpeg_Thumbnail_Quality ";
          $Shellout .= " $Output_Directory" . $Subdirectory . $Stripped_File_Name . ".ppm ";
          $Shellout .= " $Thumbs_Directory" . $Subdirectory . "t_" . $Stripped_File_Name . ".jpg";
        }
        else {
          $Shellout .= " -dither " . "-colors $Gif_Colors ";
          $Shellout .= " $Output_Directory" . $Subdirectory . $Stripped_File_Name . ".ppm ";
          $Shellout .= " $Thumbs_Directory" . $Subdirectory . "t_" . $Stripped_File_Name . ".gif";
        }
      
       if ($Verbose_Level) { print "Creating a $Thumbnail_Size $Supported_Formats[$Image_Format] thumbnail...\n"; }
    
       if ($Test_Only) { if ($Verbose_Level == 2) {print $Shellout, "\n"; } }
       else { system ($Shellout); }
    
       if ($Verbose_Level) { print "\n"; }
    
      }
      
      if ( $Clean && ($Convert_Image || $Make_Thumbnail) ) {
        print "Deleting raw output file ...\n";
        $Shellout = "rm " . $Output_Directory . $Subdirectory . $Stripped_File_Name . ".ppm ";
        if ($Test_Only) { if ($Verbose_Level == 2) {print $Shellout, "\n"; } }
        else { system ($Shellout); }
        print "\n\n";
      }
    }
  }
}